# -*- coding: utf-8 -*-
"""
Created on Sun Sep 18 10:44:11 2022

@author: richiebao caDesign设计(cadesign.cn)
"""
from Grasshopper import DataTree
from Grasshopper.Kernel.Data import GH_Path
import rhinoscriptsyntax as rs
import Rhino.Geometry as rg
import ghpythonlib.components as ghc
import ghpythonlib.treehelpers as th
import random
import itertools

def random_pts(pt,range_min,range_max,number):
    '''
    给定定位点，区间和数量，生成随机点

    Parameters
    ----------
    pt : rs(point<guid>)
        定位点.
    range_min : float
        区间控制最小值.
    range_max : float
        区间控制最大值.
    num : float/int
        随机点数量.

    Returns
    -------
    pts : rs(list<point>)
        随机点.

    '''
    x,y,z=rs.PointCoordinates(pt)
    random_val=lambda x:random.choice((1,-1))*random.uniform(range_min,range_max)+x
    x_delta=[random_val(x) for i in range(int(number))]
    y_delta=[random_val(y) for i in range(int(number))]
    z_delta=[random_val(z) for i in range(int(number))]
    pts=rs.AddPoints(zip(x_delta,y_delta,z_delta))
    
    return pts

def random_shortestLines(pts):    
    '''
    随机两两最近点连线

    Parameters
    ----------
    pts : rs(points)
        随机点.

    Yields
    ------
    rs(list<curve>)
        连线列表.

    '''
    while len(pts)>2:
        pt=pts.pop()
        closest_pt_returns=rs.PointClosestObject(pt,pts)
        pts.remove(closest_pt_returns[0])
        
        yield rs.AddLine(pt,closest_pt_returns[1])
        
def random_radius_pipe(crvs,radius_amplitude=1):
    '''
    随机管道

    Parameters
    ----------
    crvs : rs(list<curve>)
        直线列表.
    radius_amplitude : float, optional
        随机管道半径倍数. The default is 1.

    Returns
    -------
    pipes : rs(list<pipe>)
        管道列表.

    '''
    crvs_lst=list(crvs)
    n=len(list(crvs_lst))
    addPipe=lambda args:rs.AddPipe(args[0],args[1],args[2],cap=1)[0]
    radii=[random.random()*radius_amplitude for i in range(n)]
    pipes=map(addPipe,zip(crvs_lst,[0]*n,radii))
    print(pipes[0])
    
    return pipes
        
def spiral(x,y,rotation_interval):
    '''
    给定x,y序列值，和曲线旋转幅度，绘制螺旋曲线

    Parameters
    ----------
    x : list(numerical)
        x序列值.
    y : list(numerical)
        y序列值.
    rotation_interval : numerical
        曲线旋转幅度.

    Returns
    -------
    spiral_crv : rg.NurbsCurve
        螺旋曲线.
    spiral_pts : rg.Point3d
        螺旋曲线点.

    '''
    pts=[rg.Point3d(x,y,0) for x,y in zip(x,y)]
    axis_line=rg.Line(pts[-1],pts[0])
    axis_vector=pts[-1]-pts[0]
    angle_radians=0
    i=0
    spiral_pts=[]
    for pt in pts:
        closestPt=axis_line.ClosestPoint(pt,False)
        rotation=rg.Transform.Rotation(angle_radians,axis_vector,closestPt)
        angle_radians+=rotation_interval
        pt.Transform(rotation)
        spiral_pts.append(pt)
    spiral_crv=rg.NurbsCurve.Create(False,3,spiral_pts)
    return spiral_crv,spiral_pts

def space_truss_cube_pts(pt_s,pt_e,Distance,Count,):
    '''
    建立空间立方体结构随机变化点

    Parameters
    ----------
    pt_s : Point3D
        位置点（方向起始点）.
    pt_e : Point3D
        方向结束点.
    Distance : float
        点间距离.
    Count : int
        点数量（一边）.

    Returns
    -------
    random_pts_tree : Tree(Point3D)
        空间立方体结构随机变化点.

    '''
    #A-X向点（单排）
    vector_x=ghc.Vector2Pt(pt_s,pt_e,True)[0]
    vectors4Xmove=[ghc.Amplitude(vector_x,Distance*i) for i in range(Count)]
    pts_x=[ghc.Move(pt_s,v)[0] for v in vectors4Xmove]
    
    #B-Y向点（多排）
    vector_y=ghc.Vector2Pt(pt_s,ghc.Rotate(pt_e,ghc.Radians(Rotate),ghc.XYPlane(pt_s))[0],True)[0]
    vectors4Ymove=[ghc.Amplitude(vector_y,Distance*i) for i in range(Count)]
    pts_xy=[[ghc.Move(pt,v)[0] for v in vectors4Ymove] for pt in pts_x]
    
    #C-Z向点（三维矩阵）
    vector_z=ghc.UnitZ(Distance)
    vectors4Zmove=[ghc.Amplitude(vector_z,Distance*i) for i in range(Count)]
    pts_xyz=[[[ghc.Move(pt_y,v) for v in vectors4Zmove] for pt_y in pts_x] for pts_x in pts_xy]
    pts_xyz_tree=th.list_to_tree(pts_xyz)
    
    #D-随机半径球体上随机点
    random_radius=map(lambda x:x*Distance/2,ghc.Random(ghc.ConstructDomain(0,1),len(flatten_lst(pts_xyz)),Ratio_seed))
    random_sphere=[ghc.Sphere(ghc.XYPlane(pt),r) for pt,r in zip([p[0] for p in pts_xyz_tree.Branches],random_radius)]
    random_pts=[ghc.PopulateGeometry(pt,1,random.uniform(0,10000)) for pt in random_sphere]
    
    #E-转换为Tree类型数据结构
    random_pts_tree=DataTree[rg.Point3d]()
    for pt,path in zip(random_pts,pts_xyz_tree.Paths):
        random_pts_tree.Add(pt,path)
    
    return random_pts_tree
    
def space_truss_cube_crvs(pts_tree):
    '''
    建立空间立方体结构随机变化曲线

    Parameters
    ----------
    pts_tree : Tree(Point3D)
        空间立方体结构随机变化点.

    Returns
    -------
    crvs_xyz_tree : Tree(Polyline)
        空间立方体结构随机变化曲线.

    '''
    #A-y和z向折线。通过路径组织分组点
    paths=pts_tree.Paths  
    paths_idx_z=zip(['%s%s%s'%(p[0],p[1],p[2]) for p in paths],paths)
    paths_idx_y=zip(['%s%s%s'%(p[0],p[1],p[3]) for p in paths],paths)   
    crvs_xyz=[]
    
    for paths_idx in([paths_idx_z,paths_idx_y]):#
        paths_idx_sorted=sorted(paths_idx,key=lambda x:x[0][-1]) #按路径最后分枝排序，用于分组
        key_func=lambda x:x[0]            
        paths_group=itertools.groupby(paths_idx_sorted,key_func)
        paths_group_dict={k:[i[1] for i in list(v)] for k,v in paths_group}
        #print(paths_group_dict)
        crvs=[]
        for i in paths_group_dict.values():
            pts=[pts_tree.Branch(p)[0] for p in i]
            #print(pts)
            crv=ghc.PolyLine(pts,False)
            #break
            crvs.append(crv)
        crvs_xyz.append(crvs)
        
    #B-x向折线。方法1，修改path，如下述代码；方法2，用ghc.trees.TrimTree(pts_tree,2)
    paths_x=[GH_Path(p[0],p[1]) for p in paths] 
    pts_tree_x=DataTree[rg.Point3d]()
    for pt,path in zip(pts_tree.Branches,paths_x):
        pts_tree_x.Add(pt[0],path)
    pts_tree_x_flipMatrix=ghc.trees.FlipMatrix(pts_tree_x)
    crvs=ghc.PolyLine(pts_tree_x_flipMatrix,False)
    crvs_xyz.append(crvs)
    
    crvs_xyz_tree=th.list_to_tree(crvs_xyz)
    return crvs_xyz_tree

def delaunay_mesh(pts_A,pts_B,pg_count,pg_seed):
    '''
    根据两根曲线构建曲面后生成随机点建立mesh（3边面）对象

    Parameters
    ----------
    pts_A : list(Point3D)
        用于构建一根曲线的点组A.
    pts_B : list(Point3D)
        用于构建一根曲线的点组B.
    pg_count : int
        随机生成点的数量.
    pg_seed : int
        控制点位置的随机种子.

    Returns
    -------
    ptsOnsrf : list(Point3D)
        曲面上生成的随机点.
    srf : surface
        曲面.
    delaunay_mesh : mesh
        三角面.
    meshes : meshes
       拆解独立的三角面.

    '''
    crv_A=ghc.NurbsCurve(pts_A,3,False)[0]
    crv_B=ghc.NurbsCurve(pts_B,3,False)[0]
    crvs=[crv_A,crv_B]
    srf=ghc.RuledSurface(crv_A,crv_B)    
    ptsOnsrf=ghc.PopulateGeometry(srf,pg_count,pg_seed)
    ref_plane=ghc.PlaneFit(pts_A+pts_B)[0]
    delaunay_mesh=ghc.DelaunayMesh(ptsOnsrf,ref_plane)        
    deconstruct_mesh=ghc.DeconstructMesh(delaunay_mesh)
    dm_vertices=deconstruct_mesh['vertices']
    dm_faces=deconstruct_mesh['faces']
    meshes=map(lambda face:ghc.ConstructMesh(dm_vertices,face),dm_faces)

    return ptsOnsrf,srf,delaunay_mesh,meshes
    
def sorting_closestPts(pts_1,pts_2):
    '''
    根据最近点，由一组列表点（pts_1）排序另一组列表点(pts_2)

    Parameters
    ----------
    pts_1 : list(Point3D)
        参考点列表.
    pts_2 : list(Point3D)
        排序点列表.

    Returns
    -------
    sorted_pts_2 : list(Point3D)
        排序后的点列表.

    '''
    sorted_pts_2=[]
    for pt in pts_1:
        cp_return=ghc.ClosestPoint(pt,pts_2)
        cp=cp_return['closest_point']
        cp_idx=cp_return['cp_index']
        sorted_pts_2.append(pts_2.pop(cp_idx))
        
    return sorted_pts_2

def wallOpening_typeA(args):
    '''
    墙体（幕墙）开洞模式-类型A

    Parameters
    ----------
    args : mesh(一个三边面)，float
        mesh,distance.

    Returns
    -------
    srfs : list(surface)
        开洞列表.

    '''
    mesh,distance=args
    fb=ghc.FaceBoundaries(mesh)
    pts=ghc.PopulateGeometry(mesh,3,random.uniform(0,10000))
    triangle_pl=ghc.PolyLine(pts,True)
    triangle_srf=ghc.BoundarySurfaces(triangle_pl)
    srfs=[]
    crvs=[]
    for dis in [distance,ghc.Negative(distance)]:
        offset_srf=ghc.OffsetSurfaceLoose(triangle_srf,dis,True)
        naked_edges=ghc.JoinCurves(ghc.BrepEdges(offset_srf)['naked'],True)
        naked_pts=ghc.ControlPoints(naked_edges)['points'][:3]
        fb_pts=ghc.ControlPoints(fb)['points'][:3]
        naked_pts_sorted=sorting_closestPts(fb_pts,naked_pts)
        crv_1=ghc.PolyLine(fb_pts,True)
        crv_2=ghc.PolyLine(naked_pts_sorted,True)
        ruledSrf=ghc.RuledSurface(crv_1,crv_2)
        srfs.append(ruledSrf)
        crvs.append(crv_2)
    ruledSrf_crvs=ghc.RuledSurface(crvs[0],crvs[1])
    srfs.append(ruledSrf_crvs)
    return srfs